from airflow.hooks.base import BaseHook
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.context import Context

from airflow_tools.filesystems.filesystem_factory import FilesystemFactory


class FilesystemFileSensor(BaseSensorOperator):
    """
    Check if a file exists in the specified storage.

    This sensor use connection type (`connection.conn_type`) to determine the
    filesystem type.

    Current supported filesystems:

    - 'aws' (S3)
    - 'azure_databricks_volume'
    - 'azure_file_share_sp'
    - 'sftp'
    - 'fs' (LocalFileSystem)
    - 'google_cloud_platform'
    - 'wasb' (BlobStorageFilesystem)

    :param filesystem_conn_id: connection id for the filesystem.
    :param source_path: The path to the file (include bucket name).

    """

    template_fields = ('source_path',)

    def __init__(self, filesystem_conn_id: str, source_path: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.source_path = source_path
        self.filesystem_conn_id = filesystem_conn_id

    def poke(self, context: Context):
        """
        Check if the file exists in the specified storage.
        """
        conn = BaseHook.get_connection(self.filesystem_conn_id)
        fs = FilesystemFactory.get_data_lake_filesystem(conn)

        return fs.check_file(self.source_path)



from airflow.hooks.base import BaseHook
from azure.identity import ClientSecretCredential


class AzureFileShareServicePrincipalHook(BaseHook):
    """
    Requires defined connection with this structure:
        conn_type: azure_file_share_sp
        host: <hostname>
        login: <service_principal_id>
        password: <service_principal_secret>
        extra: {"tenant_id": "<tenant_id>", "share_name": "<share_name>", "protocol": "https"}
    """

    def __init__(self, conn_id: str):
        super().__init__()
        self.conn_id = conn_id
        self.connection = self.get_connection(conn_id)
        self.host = self.connection.host
        self.service_principal_id = self.connection.login
        self.service_principal_secret = self.connection.password
        self.tenant_id = self.connection.extra_dejson.get("tenant_id")
        self.share_name = self.connection.extra_dejson.get("share_name")
        self.protocol = self.connection.extra_dejson.get("protocol", "https")

        self.credentials = self.get_token_credentials()
        self.account_url = f"{self.protocol}://{self.host}"

    def get_token_credentials(self):
        return ClientSecretCredential(
            client_id=self.service_principal_id,
            client_secret=self.service_principal_secret,
            tenant_id=self.tenant_id,
        )

    def get_conn(self):
        from azure.storage.fileshare import ShareClient

        return ShareClient(
            account_url=self.account_url,
            credential=self.credentials,
            share_name=self.share_name,
            token_intent='backup',
        )



import io
import logging
import sys
import typing

import pandas as pd
from airflow.hooks.base import BaseHook
from airflow.models import BaseOperator
from sqlalchemy import create_engine, inspect, engine, Integer, Float, String, DateTime, Boolean

from airflow_tools.filesystems.filesystem_factory import FilesystemFactory

if sys.version_info < (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

logger = logging.getLogger(__name__)


type_mapping = {
    'int64': Integer, 'int': Integer, 'integer': Integer,
    'float64': Float, 'float': Float, 
    'object': String, 'string': String, 'str': String,
    'datetime64[ns]': DateTime,
    'bool': Boolean, 'boolean': Boolean,
    
}

class FilesystemToDatabaseOperator(BaseOperator):
    """
    This operator will copy a file from a filesystem to a database table.
    """

    template_fields = (
        'db_table',
        'db_schema',
        'filesystem_conn_id',
        'database_conn_id',
        'filesystem_path',
        'metadata',
    )

    def __init__(
        self,
        filesystem_conn_id: str,
        database_conn_id: str,
        filesystem_path: str,
        db_table: str,
        db_schema: typing.Optional[str] = None,
        source_format: typing.Optional[Literal['csv', 'json', 'parquet']] = 'csv',
        source_format_options: typing.Optional[typing.Dict] = None,
        table_aggregation_type: typing.Optional[
            Literal['append', 'fail', 'replace']
        ] = 'append',
        metadata: typing.Optional[typing.Dict[str, str]] = None,
        include_source_path: bool = True,
        *args,
        **kwargs,
    ) -> None:

        super().__init__(*args, **kwargs)

        self.filesystem_conn_id = filesystem_conn_id
        self.database_conn_id = database_conn_id
        self.filesystem_path = filesystem_path
        self.db_table = db_table
        self.db_schema = db_schema
        self.source_format = source_format
        self.source_format_options = source_format_options
        self.table_aggregation_type = table_aggregation_type
        self.metadata = metadata or {'_DS': '{{ ds }}'}
        self.include_source_path = include_source_path

    def execute(self, context):
        logger.info(f'Create connection for filesystem ({self.filesystem_conn_id})')
        filesystem = FilesystemFactory.get_data_lake_filesystem(
            connection=BaseHook.get_connection(self.filesystem_conn_id),
        )

        logger.info(f'Create SQLAlchemy engine with connection_id {self.database_conn_id}')
        engine = create_engine(
            BaseHook.get_connection(self.database_conn_id).get_hook().get_uri()
        )
        
        for blob_path in filesystem.list_files(prefix=self.filesystem_path):

            if not blob_path.endswith(
                (f'.{self.source_format}', f'.{self.source_format}.gz')
            ):
                logger.warning(f'Blob {blob_path} is not in the right format. Skipping...')
                continue

            logger.info(f'Read file {blob_path} and convert to pandas')
            raw_content = io.BytesIO(filesystem.read(blob_path))

            df = self.raw_content_to_pandas(path_or_buf=raw_content)
            
            self._check_and_fix_column_differences(df, self.db_table, engine)   

            for key, value in self.metadata.items():
                df[key] = value
                
                if key.startswith('_'):
                    # _* fields are treated as metadata and we try to convert them in datetimes
                    df[key] = self._convert_to_datetime(df[key])
                    
            if self.include_source_path:
                df['_LOADED_FROM'] = blob_path
                df['_LOADED_FROM'] = df['_LOADED_FROM'].astype('string')

            df.to_sql(
                name=self.db_table,
                schema=self.db_schema,
                con=engine,
                if_exists=self.table_aggregation_type,
                index=False,
            )
            
    def _check_and_fix_column_differences(self, df: pd.DataFrame, table_name: str, engine: engine.Engine):
        """
        This method checks if the columns in the dataframe are the same as the
        ones in the database table. If they are not, it adds the missing 
        columns to the database table or fills them with None.

        Args:
            df (pd.DataFrame): dataframe to be inserted into the database
            table_name (str): name of the database table
            engine (engine.Engine): SQLAlchemy engine object for the database connection
        """
        inspector = inspect(engine)

        if table_name in inspector.get_table_names():
            source_file_columns = set(df.columns.tolist())
            table_columns = set([col['name'] for col in inspector.get_columns(self.db_table) if col['name'] not in self.metadata])
            
            # Column in source file but not present in db table
            only_csv_columns = source_file_columns - table_columns          
            with engine.connect() as conn:
                for column in only_csv_columns:
                    logger.warning(
                        f'Table "{table_name}" does not have column "{column}" present in the source file, '
                        f'column "{column}" will be filled with null values.'
                    )
                    sqlalchemy_type = type_mapping.get(df.dtypes.get(column), String)
                    conn.execute(f'ALTER TABLE {table_name} ADD COLUMN "{column}" {sqlalchemy_type.__name__}')
            
            # Column in db table but not present in the source file
            only_db_table_columns = table_columns - source_file_columns
            for column in only_db_table_columns:
                logger.warning(
                    f'Source file does not have column "{column}" present in the table "{table_name}", '
                    f'column "{column}" will be filled with null values.'
                )
                df[column] = None          
        
    @staticmethod
    def _convert_to_datetime(value):
        try:
            return pd.to_datetime(value)
        except ValueError:
            return value

    def raw_content_to_pandas(self, path_or_buf: typing.Union[str, bytes, io.StringIO]):
        options = self.source_format_options or {}

        match self.source_format:
            case 'csv':
                return pd.read_csv(path_or_buf, **options)
            case 'json':
                return pd.read_json(path_or_buf, **options)
            case 'parquet':
                return pd.read_parquet(path_or_buf, **options)
            case _:
                raise ValueError(f'Unknown source format {self.source_format}')



import os
import typing

from airflow.configuration import conf
from airflow.providers.slack.notifications.slack_webhook import (
    send_slack_webhook_notification,
)

SLACK_WEBHOOK_CONN = 'slack_webhook_notification_conn'
DEFAULT_DAG_MSG = ':red_circle: *DAG _\"{{dag.dag_id}}\"_ failed*'
DEFAULT_TASK_MSG = ':red_circle: *Task _\"{{task.task_id}}\"_ failed*'
SLACK_NOTIFICATION_IMG_URL = os.getenv('AIRFLOW_TOOLS__SLACK_NOTIFICATION_IMG_URL')


class Source:
    DAG = 'DAG'
    TASK = 'TASK'


def _get_message_blocks(
    text: str, image_url: str, source: str
) -> list[dict[str, typing.Any]]:
    base_url = conf.get('webserver', 'BASE_URL')
    dag_execution_logs_url = (
        base_url + '/dags/{{ dag.dag_id }}/grid?'
        'dag_run_id={{ run_id | urlencode }}&'
        'task_id={{ task.task_id }}&tab=logs'
    )

    match source:
        case Source.DAG:
            failed_tasks = "{% for t in dag_run.get_task_instances(state='failed') %}'{{ t.task_id }}', {% endfor %}"
        case Source.TASK:
            failed_tasks = "{{ task.task_id }}"
        case _:
            failed_tasks = '"unknown source"'

    blocks = [
        {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": str(text) + ":\n"
                "```"
                "> Run_id:          {{ run_id }}\n"
                "> Execution_date:  {{ dag_run.execution_date }}\n"
                "> Queued_at:       {{ dag_run.queued_at }}\n"
                "> Failed tasks:    [" + failed_tasks + "]\n"
                "> Host:            " + str(base_url) + "  \n"
                "```",
            },
        },
        {
            "type": "section",
            "text": {"type": "mrkdwn", "text": "*More info (logs):*"},
            "accessory": {
                "type": "button",
                "text": {
                    "type": "plain_text",
                    "text": "Check in airflow webserver",
                    "emoji": True,
                },
                "value": "airflow",
                "url": str(dag_execution_logs_url),
                "action_id": "button-action",
            },
        },
        {"type": "divider"},
    ]

    if image_url:
        blocks[0]["accessory"] = {
            "type": "image",
            "image_url": str(image_url),
            "alt_text": "Airflow notification (via Slack incoming webhook)",
        }

    return blocks


def dag_failure_slack_notification_webhook(
    conn_id: str = SLACK_WEBHOOK_CONN,
    text: str = '',
    blocks: typing.List[dict[str, typing.Any]] | None = None,
    include_blocks: bool = True,
    source: typing.Literal['DAG', 'TASK'] = Source.DAG,
    image_url: str | None = SLACK_NOTIFICATION_IMG_URL,
) -> typing.Callable:

    """
    Sends a Slack notification for DAG failure using a webhook.

    Args:
        conn_id (str) [optional]: The connection ID for the Slack webhook.
        text (str) [optional]: The text message to be sent in the Slack notification. Default:
        blocks (dict) [optional]: The blocks to be included in the Slack notification.
        include_blocks (bool) [optional]: Whether to include default message blocks if `blocks` is not provided.
        source (typing.Literal['DAG', 'TASK']) [optional]: The source of the failure (DAG or TASK). Default: 'DAG'.
        image_url (str) [optional]: The URL of the image to be included in the Slack notification.

    Returns:
        typing.Callable: A callable object that sends the Slack notification.

    """

    if not text:
        text = DEFAULT_DAG_MSG if source == Source.DAG else DEFAULT_TASK_MSG

    if not blocks and include_blocks:
        blocks = _get_message_blocks(text=text, image_url=image_url, source=source)

    return send_slack_webhook_notification(
        slack_webhook_conn_id=conn_id, text=text, blocks=blocks
    )



import logging
from io import BytesIO

from azure.storage.fileshare._models import DirectoryProperties

from airflow_tools.filesystems.filesystem_protocol import FilesystemProtocol
from airflow_tools.providers.azure.hooks.azure_file_share import (
    AzureFileShareServicePrincipalHook,
)

logger = logging.getLogger(__file__)


class AzureFileShareFilesystem(FilesystemProtocol):
    def __init__(self, hook: AzureFileShareServicePrincipalHook):
        self.hook = hook

    def read(self, path: str) -> bytes:
        return self.hook.get_conn().get_file_client(path).download_file().readall()

    def write(self, data: str | bytes | BytesIO, path: str):
        if isinstance(data, str):
            data = data.encode()
        elif isinstance(data, BytesIO):
            data = data.getvalue()
        self.hook.get_conn().get_file_client(path).upload_file(data)

    def delete_file(self, path: str):
        self.hook.get_conn().get_file_client(path).delete_file()

    def create_prefix(self, prefix: str):
        self.hook.get_conn().get_directory_client(prefix).create_directory()

    def delete_prefix(self, prefix: str):
        conn = self.hook.get_conn()
        for item in conn.list_directories_and_files(prefix):
            if isinstance(item, DirectoryProperties):
                self.delete_prefix(f'{prefix}/{item.name}')
            else:
                conn.get_file_client(f'{prefix}/{item.name}').delete_file()
        conn.get_directory_client(prefix).delete_directory()

    def check_file(self, path: str) -> bool:
        return self.hook.get_conn().get_file_client(path).exists()

    def check_prefix(self, prefix: str) -> bool:
        return self.hook.get_conn().get_directory_client(prefix).exists()

    def list_files(self, prefix: str) -> list[str]:
        return [
            f'{prefix}/{item.name}'
            for item in self.hook.get_conn().list_directories_and_files(prefix)
            if not item.is_directory
        ]



import logging

from airflow.decorators import branch_task
from airflow.hooks.base import BaseHook

from airflow_tools.filesystems.filesystem_factory import FilesystemFactory

logger = logging.getLogger(__file__)


@branch_task
def branch_filesystem_check_task(
    filesystem_conn_id: str,
    filesystem_path: str,
    main_branch: str,
    alternative_branch: str,
) -> bool:
    filesystem_protocol = FilesystemFactory.get_data_lake_filesystem(
        connection=BaseHook.get_connection(filesystem_conn_id),
    )
    logger.info(f"Trying to check: {filesystem_path}")
    if (
        filesystem_protocol.check_prefix(filesystem_path)
        if filesystem_path.endswith("/")
        else filesystem_protocol.check_file(filesystem_path)
    ):
        return main_branch
    else:
        return alternative_branch



import logging
from io import BytesIO

from databricks.sdk.errors.platform import NotFound

from airflow_tools.filesystems.filesystem_protocol import FilesystemProtocol
from airflow_tools.providers.azure.hooks.azure_databricks import (
    AzureDatabricksVolumeHook,
)

logger = logging.getLogger(__file__)


class AzureDatabricksVolumeFilesystem(FilesystemProtocol):
    def __init__(self, hook: AzureDatabricksVolumeHook):
        self.hook = hook

    def read(self, path: str) -> bytes:
        return self.hook.get_conn().files.download(path).contents

    def write(self, data: str | bytes | BytesIO, path: str):
        if isinstance(data, str):
            data = data.encode()
        elif isinstance(data, BytesIO):
            data = data.getvalue()
        self.hook.get_conn().files.upload(path, data)

    def delete_file(self, path: str):
        self.hook.get_conn().files.delete(path)

    def create_prefix(self, prefix: str):
        self.hook.get_conn().files.create_directory(prefix)

    def delete_prefix(self, prefix: str):
        conn = self.hook.get_conn()

        try:
            entries = list(conn.files.list_directory_contents(prefix))
        except NotFound:
            return

        for entry in entries:
            if entry.is_directory:
                self.delete_prefix(entry.path)
            else:
                conn.files.delete(entry.path)
        conn.files.delete_directory(prefix)

    def check_file(self, path: str) -> bool:
        prefix = path.rsplit("/", 1)[0]

        try:
            file_list_path = [
                entry.path == path
                for entry in self.hook.get_conn().files.list_directory_contents(prefix)
                if not entry.is_directory
            ]
            return any(file_list_path)

        except NotFound:
            return False

    def check_prefix(self, prefix: str) -> bool:
        try:
            _ = [
                f
                for f in self.hook.get_conn().files.list_directory_contents(
                    prefix, page_size=1
                )
            ]
            return True

        except NotFound:
            return False

    def list_files(self, prefix: str) -> list[str]:
        return [
            entry.path
            for entry in self.hook.get_conn().files.list_directory_contents(prefix)
            if not entry.is_directory
        ]



import logging
from io import BytesIO

from airflow.providers.microsoft.azure.hooks.wasb import WasbHook

from airflow_tools.filesystems.filesystem_protocol import FilesystemProtocol

logger = logging.getLogger(__file__)


class BlobStorageFilesystem(FilesystemProtocol):
    def __init__(self, hook: WasbHook):
        self.hook = hook

    def read(self, path: str) -> bytes:
        container_name, blob_name = _get_container_and_blob_name(path)
        stream = self.hook.download(container_name=container_name, blob_name=blob_name)
        return stream.readall().encode()

    def write(self, data: str | bytes | BytesIO, path: str):
        container_name, blob_name = _get_container_and_blob_name(path)
        if isinstance(data, str):
            self.hook.load_string(data, container_name, blob_name)
            return

        if isinstance(data, BytesIO):
            data = data.getvalue()
        logger.info(
            f'Writing to wasb container "{container_name}" and blob "{blob_name}"'
        )
        self.hook.upload(container_name=container_name, blob_name=blob_name, data=data)

    def delete_file(self, path: str):
        container_name, blob_name = _get_container_and_blob_name(path)
        self.hook.delete_file(
            container_name, blob_name, is_prefix=False, ignore_if_missing=True
        )

    def create_prefix(self, prefix: str):
        container_name, blob_name = _get_container_and_blob_name(f'{prefix}/')
        self.hook.upload(container_name, blob_name, data=b"")

    def delete_prefix(self, prefix: str):
        container_name, blob_prefix = _get_container_and_blob_name(prefix)
        self.hook.delete_file(
            container_name, blob_prefix, is_prefix=True, ignore_if_missing=True
        )

    def check_file(self, path: str) -> bool:
        container_name, blob_name = _get_container_and_blob_name(path)
        return self.hook.check_for_blob(container_name, blob_name)

    def check_prefix(self, prefix: str) -> bool:
        container_name, blob_prefix = _get_container_and_blob_name(prefix)
        return self.hook.check_for_prefix(container_name, blob_prefix)

    def list_files(self, prefix: str) -> list[str]:
        container_name, blob_prefix = _get_container_and_blob_name(prefix)
        return self.hook.get_blobs_list(container_name, blob_prefix)


def _get_container_and_blob_name(path: str) -> tuple[str, str]:
    parts = path.split("/")
    return parts[0], "/".join(parts[1:])



from io import BytesIO

from airflow.providers.google.cloud.hooks.gcs import GCSHook

from airflow_tools.filesystems.filesystem_protocol import FilesystemProtocol


class GCSFilesystem(FilesystemProtocol):
    def __init__(self, hook: GCSHook):
        self.hook = hook

    def read(self, path: str) -> bytes:
        bucket_name, object_name = _get_bucket_and_key_name(path)
        return self.hook.download(bucket_name, object_name)

    def write(self, data: str | bytes | BytesIO, path: str):
        bucket_name, object_name = _get_bucket_and_key_name(path)
        self.hook.upload(bucket_name, object_name, data=data)

    def delete_file(self, path: str):
        bucket_name, object_name = _get_bucket_and_key_name(path)
        self.hook.delete(bucket_name, object_name)

    def create_prefix(self, prefix: str):
        bucket_name, key_prefix = _get_bucket_and_key_name(prefix)
        if not self.hook.bucket_exists(bucket_name):
            self.hook.create_bucket(bucket_name)
        # Create an empty object to represent the prefix
        self.hook.upload(bucket_name, key_prefix, data='')

    def delete_prefix(self, prefix: str):
        bucket_name, key_prefix = _get_bucket_and_key_name(prefix)
        objects = self.hook.list(bucket_name, prefix=key_prefix)
        if objects:
            for object_name in objects:
                self.hook.delete(bucket_name, object_name)

    def check_file(self, path: str) -> bool:
        bucket_name, object_name = _get_bucket_and_key_name(path)
        return self.hook.exists(bucket_name, object_name)

    def check_prefix(self, prefix: str) -> bool:
        bucket_name, key_prefix = _get_bucket_and_key_name(prefix)
        objects = self.hook.list(bucket_name, prefix=key_prefix, max_results=1)
        return bool(objects)

    def list_files(self, prefix: str) -> list[str]:
        bucket_name, key_prefix = _get_bucket_and_key_name(prefix)
        objects = self.hook.list(bucket_name, prefix=key_prefix)
        return [f'{bucket_name}/{object_path}' for object_path in (objects or [])]


def _get_bucket_and_key_name(path: str) -> tuple[str, str]:
    parts = path.split("/")
    return parts[0], "/".join(parts[1:])
